Migrating from vObject 2.0 to 3.0
=================================

vObject 3.0 got a major overhaul, and much better built-in support for all
kinds of properties and escaping.

This version fixes the most important bugs, specifically Issue #19.

To do this, a few backwards compatibility breaks had to be made. This document
describes each of them, as well as all the new features.

New features
------------

### Serializer now properly deals with escaped commas and semi-colons.

When a semi-colon was embedded in many properties as `\;`, VObject would escape
this as `\\;`, while this was not the intention.

Now there's a distinction between single-value properties and multi-value
properties.

Example:
```php
<?php

$card->ORG = ['Company', 'Department'];

echo (string)$card->ORG; // Will echo "Company, Department";
print_r($card->getParts()); // Prints an array with both elements.

?>
```

The new `getParts()` method will now always return an array, regardless of
the number of elements. `getValue()` or `(string)$property` will always try
to combine the array into a string.

In addition, SabreDAV now has a record with all known property names that may
contain multiple values, and also whether they use a comma, or a semi-colon
for a separator.

If a comma or semi-colon was embedded unescaped, and it was not expected that
the property had more than one value, or the other character is used for
multiple values, the comma or semi-colon will be treated as if it were escaped.

### Easier to create Date values:

```php
<?php

// Setting datetime objects was not supported before.
$event->DTSTART = new DateTime('now');

?>
```

### Easier to create new documents.

A syntax like this is now supported to create a document:

```php
<?php

$properties = [
    'FN' => 'Evert Pot',
    'N' => ['Pot','Evert'],
    'ORG' => ['fruux GmbH', 'dev team'],
];
$card = new Sabre\VObject\Component\VCard($properties);

?>
```

iCalendar works the same way:

```php
<?php

$properties = [
    'VEVENT' => [
        'SUMMARY' => 'Daily meeting',
        'DTSTART '=> new DateTime('2013-05-13 14:00:00'),
        'RRULE' => ['freq' => 'daily', 'bydays' => '1,2,3,4,5'],
    ],
];
$cal = new Sabre\VObject\Component\VCalendar($properties);


?>
```

Properties such as `CALSCALE`, `VERSION` and `PRODID` will automatically be added.

### RFC6868 is used to serialize parameters.

[RFC6868][1] uses the caret (`^`) to escape parameter values. This may look
something like this in a vCard:

```
TEL;LABEL="The ^'old^' house":+44 555 1243253647
```

In that case, the label will now be interpreted as:

```
The "old" house
```

Both the parser and serializer now support this syntax.


### Broken vCard 2.1 generated by Microsoft are now properly supported.

Before, an `INGORE_INVALID_LINES` option had to be passed to the parser to just
get these to parse, albeit not exactly how you'd expect.

Now the parser can properly detect Microsoft's broken vCards, will
automatically repair them, and decode the `QUOTED-PRINTABLE` format.

To use this new behaviour, you must pass the `FORGIVING` option to the parser.

### The add() method now returns the object it's creating.

Using the add() method has become a lot more straightforward.

```php
<?php

$card = Sabre\VObject\Component\VCard();

$tel = $card->add('TEL', '+31 555 333444');
$tel['TYPE'] = 'FAX';

?>
```

### The parser now supports streams

This means that the entire input no longer has to be kept into memory, and you
can parse straight from files without having to load them in.

To use this, simply pass a stream instead of a filename:

```php
<?php

$stream = fopen('hugecalendar.ics','r');
$calendar = Sabre\VObject\Reader::read($stream);

?>
```

### Added a remove() method to Component

The remove method allows for much better manipulation of existing objects.
To remove all fax numbers from a vCard, the following syntax may be used:

```php
<?php
foreach($card->TEL as $tel) {

    if ((string)$tel->TYPE === 'FAX') {
        $card->remove($tel);
    }

}
?>
```

Aside from removing objects (`$tel` in the prevous example), you can also just
specify a property name, in which case all properties with that name will be
stripped.

### Every property, parameter and component has a reference to the document.

By calling `$property->root`, you can easily obtain a reference to the top-level
component.

### Documents have a getDocumentType method.

The 2 document classes (`Component\VCalendar` and `Component\VCard`) have a
`getDocumentType()` method, which return one of the following constants:

* `Document::ICALENDAR20`
* `Document::VCARD21`
* `Document::VCARD30`
* `Document::VCARD40`
* `Document::UNKNOWN`

### Binary properties are automatically decoded.

Properties that contain binaries, such as `ATTACH` and `PHOTO` generally have
their data encoded as base64. This is now automatically encoded and decoded.

Just set the file and you're good:

```php
$vcard->PHOTO = file_get_contents('yourface.jpg');
```

Backwards compatibility breaks
------------------------------

### Creating components

Before, it was possible to create components such as `VEVENT`, `VTODO`, etc
with this syntax:

```php
<?php

$event = \Sabre\VObject\Component::create('VEVENT');
$event->summary = 'Birthday party!';

// Or:

$event = new \Sabre\VObject\Component('VEVENT');
$event->summary = 'Birthday party!';

?>
```

Neither of those are legal any longer. Components now _must_ be created through
the Document object.

Example:

```php
<?php

    $vcalendar = new \Sabre\VObject\Component\VCalendar();
    $event = $vcalendar->createComponent('VEVENT');
    $event->summary = 'Birthday party!';

?>
```

Or:

```php
$vcalendar = new \Sabre\VObject\Component\VCalendar();
$vcalendar->add('VEVENT', [
    'summary' => 'Birthday party!',
]);
```

### Creating properties

Creating properties works the _exact_ same way.

Old:

```php
<?php

$location = new \Sabre\VObject\Property('LOCATION', 'Home');

// Or:

$location = \Sabre\VObject\Property::create('LOCATION', 'Home');

?>
```

Now you must also use the document:

Example:

```php
<?php

$card = new \Sabre\VObject\Component\VCard();
$location = $card->createProperty('LOCATION','Home');

?>
```

Note that in most cases, this syntax is highly recommended instead:

```php
<?php

$card = new \Sabre\VObject\Component\VCard();
$location = $card->add('LOCATION','Home');

?>
```

In this case it doesn't make much of a difference, but when constructing
highly complex objects with sub-components, this _will_ make a big difference.

### Component::children() and Property::parameters() return arrays.

Before vObject 3 they returned an `ElementList`.

### The signature for DateTime::setDateTime has changed.

Before, you would use the following 4 syntaxes to set the date and time:

```php
<?php

use Sabre\VObject\Property;

$now = new DateTime('now');

$dt = new Property\DateTime('DTSTART');
$dt->setDateTime($now, Property\DateTime::DATE); // Date only
$dt->setDateTime($now, Property\DateTime::LOCAL); // Floating time
$dt->setDateTime($now, Property\DateTime::UTC); // Convert to UTC
$dt->setDateTime($now, Property\DateTime::LOCALTZ); // Local to timezone information.

```

This has completely changed:

```
<?php

// Date only
$now = new DateTime('now');
$dt = $calendar->create('DTSTART');
$dt->setValue($now);
$dt['VALUE'] = 'date';

// Floating time
$dt = $calendar->create('DTSTART');
$dt->setValue($now, $floating = true);

// Convert to UTC
$now->setTimeZone(\DateTimeZone('UTC'));
$dt = $calendar->create('DTSTART');
$dt->setValue($now);

// Local time + timezone information
$dt = $calendar->create('DTSTART');
$dt->setValue($now);

?>
```

Note that the preceeding examples are all a bit convoluted.
In most cases you just want to do something like:

```php
<?php

// Assuming this a vevent
$event->DTSTART = $now;

?>
```

In addition, the `MultiDateTime` property is no more, and it's methods are
simply merged into `DateTime`.

### The $value property is now protected everywhere.

Both the `Property` and the `Parameter` classes had a public `$value` property,
which allowed you to retrieve the string value for either of those.

This is now protected, so you must access it in this manner:

```php
<?php
// Assuming $prop is a property object.

$prop->setValue('Birthday');
echo $prop->getValue();

?>
```

// For properties that have more than 1 value, you can use `setParts` and
`getParts`:


```php
<?php

$org->setParts(['Company', 'Department']);
print_r($org->getParts());

?>
```

### Binary properties are automatically de- and encoded.

The `ATTACH`, `LOGO` and `PHOTO` properties now automatically de- and encode
their binary values. In vObject 2 they were accessed by their raw base64
values.

### Components and documents get injected with default properties.

When creating a new `VCalendar`, it will automatically get the `VERSION`,
`CALSCALE` and `PRODID` properties.

If you were adding your own with this syntax:

```php
<?php

$calendar = new Sabre\VObject\Component\VCalendar();
$calendar->VERSION = '2.0';

?>
```

Then nothing will go wrong, and the properties will simply be overwritten.
However, if you used `add()` before in this manner:

```php
<?php

$calendar = new Sabre\VObject\Component\VCalendar();
$calendar->add('version', '2.0');

?>
```

You will end up with 2 VERSION properties, making the document invalid.

### componentMap and propertyMap properties have moved.

When you wanted to automatically map certain components or properties to
certain PHP classes, you could do so with `Component::$componentMap` and
`Property::$propertyMap`.

These properties have now moved to the document classes:

* `Component\VCalendar::$propertyMap`
* `Component\VCalendar::$componentMap`
* `Component\VCard::$propertyMap`
* `Component\VCard::$componentMap`

[1]: http://tools.ietf.org/html/rfc6868
